#ifndef ns_dtn_h
#define ns_dtn_h

#include "agent.h"
#include "packet.h"
#include "mac.h"
#include "mac-802_11.h"

// --------------------------------------------- DTN header -------------------------------------------------- //

#define DTN_HEADER_SIZE	50
#define DEFAULT_TIMEOUT 2

struct hdr_dtn {
	int							nSequenceNumber_;
	nsaddr_t					nSourceAddress_;
	double						fTimestamp_;
	int							nFragmentOffset_;
	int							nSegmentNumber_; // Possibly useless
	int							nBundleSize_;
	int							nADUSize_;
	int							nAvailableBufferSpace_;
	int							nPayload_;
	int							nNextRVms_;
	unsigned char				bIsStatusReport_;
	unsigned char				bCustodyAccepted_;
	unsigned char				bCustodyRejected_;
	unsigned char				bLeadingBytes_;

	static int 					offset_;	// offset for this header
	inline static int& 			offset() 				{ return offset_; }
	inline static hdr_dtn* 		access(Packet* p) 		{ return (hdr_dtn*) p->access(offset_); }
	
	/* per-field member functions */
	int& 						SequenceNumber()		{ return (nSequenceNumber_); }
	nsaddr_t& 					SourceAddress()			{ return (nSourceAddress_); }
	double& 					Timestamp()				{ return (fTimestamp_); }
	int& 						FragmentOffset()		{ return (nFragmentOffset_); }
	int& 						SegmentNumber()			{ return (nSegmentNumber_); }
	int& 						BundleSize()			{ return (nBundleSize_); }
	int& 						ADUSize()				{ return (nADUSize_); }
	int& 						AvailableBufferSpace() 	{ return (nAvailableBufferSpace_); }
	int& 						Payload()				{ return (nPayload_); }
	int& 						NextRVms()				{ return (nNextRVms_); }
	unsigned char& 				IsStatusReport()		{ return (bIsStatusReport_); }
	unsigned char& 				CustodyAccepted()		{ return (bCustodyAccepted_); }
	unsigned char& 				CustodyRejected()		{ return (bCustodyRejected_); }
	unsigned char& 				LeadingBytes()			{ return (bLeadingBytes_); }
};

// ----------------------------------------- Bundle -------------------------------------------- //

class DtnConnection;

class Bundle {
public:
	Bundle();
	~Bundle();

	void 					InitLocalVariables();
	
	hdr_dtn&				HeaderDTN()				{ return headerDTN_; }
	int&					OriginalOffset()		{ return nOriginalOffset_; }
	int&					OriginalSize()			{ return nOriginalSize_; }
	int&					TCPSequenceNo()			{ return nTCPSequenceNo_; }
	int&					ReceivedBytes()			{ return nReceivedBytes_; }
	double&					RetransmitTime()		{ return fRetransmitTime_; }
	unsigned char&			Forwarded()				{ return bForwarded_; }
	unsigned char&			Acked()					{ return bAcked_; }
	unsigned char&			Suspended()				{ return bSuspended_; }
	DtnConnection*&			DownstreamConnection()	{ return pDownstreamCon_; }
	DtnConnection*&			UpstreamConnection()	{ return pUpstreamCon_; }
	
	Bundle*&				NextBundle()			{ return pNextBundle_; }

	bool					IsSameAs(Bundle otherBundle);
	bool					IsInSameADUAs(Bundle otherBundle);
	bool					Precedes(Bundle otherBundle);
	bool					IsCompleteADU();
	int						LastBundleByte();
	int						LastReceivedByte();
	bool					CanBeMergedWith(Bundle otherBundle);
	bool					MergeWith(Bundle otherBundle);
	bool					FragmentReceivedBytes();
	bool					FragmentAt(int nOffset);
	bool					FullyContains(Bundle otherBundle);
	bool					OriginalFullyContains(Bundle otherBundle);
	bool					IsFullyReceived();
	void					RejectCustody();
	void					AcceptCustody();
	void					Dump();
	bool					IsInTransit();

protected:
	void					InitBundleVariables();

	hdr_dtn					headerDTN_;
	int						nOriginalOffset_;
	int						nOriginalSize_;
	int						nTCPSequenceNo_;
	int						nReceivedBytes_;
	double					fRetransmitTime_;
	unsigned char			bForwarded_;
	unsigned char			bAcked_;
	unsigned char			bSuspended_;
	DtnConnection*			pDownstreamCon_;
	DtnConnection*			pUpstreamCon_;

	Bundle*					pNextBundle_;
};

// ----------------------------------------- BundleManager -------------------------------------------- //

class BundleManagerDTN {

public:
	BundleManagerDTN();
	~BundleManagerDTN();
	
	Bundle*			AddBundle(Bundle newBundle); 
	Bundle*			GetFirstCompletedADU(Bundle* pAfterBundle);
	Bundle*			GetFirstUnSentBundle(Bundle* pAfterBundle);
	Bundle*			GetFirstExpiredBundle(Bundle* pAfterBundle);
	Bundle*			FindBundle(Bundle inBundle);
	void			Cleanup();
	void			EmptyList();
	void			RemovePointer(Bundle *pBundle);
	void			RemoveBundle(Bundle inBundle);
	bool			CleanIfOriginalAcked(Bundle inBundle);
	bool			OriginalBundleAcked(Bundle inBundle);
	bool			OriginalBundleCompleted(Bundle inBundle);
	int				BundlesStored();
	int				MemoryAllocated();
	void			MergeAllFragments();
	void			FragmentAllBundles();
	void			DumpBundles();
	Bundle*			FindOriginalBundle(Bundle inBundle);
	double			FindEarlierRTTime();
protected:

	Bundle			*pHead_;
};

// --------------------------------------------- DtnAgentRVTimer -------------------------------------------------- //

class DtnAgent;

class DtnAgentRVTimer : public TimerHandler {
public:
	DtnAgentRVTimer(DtnAgent* t) : TimerHandler(), t_(t) {}
	virtual void expire(Event*);
protected:
	DtnAgent* t_;
};

class DtnAgentRTTimer : public TimerHandler {
public:
	DtnAgentRTTimer(DtnAgent* t) : TimerHandler(), t_(t) {}
	virtual void expire(Event*);

	double SetoffTime();
protected:
	DtnAgent* t_;
};

// ----------------------------------------------- DtnConnection ------------------------------------------------ //

class TcpDtnAgent;
class TcpDtnSink;

class DtnConnection {
public:
	DtnConnection(int nConnectionID, DtnAgent *pOwnerDtnAgent, DtnAgent *pPeerDtnAgent);
	~DtnConnection();

	int&					ConnectionID()		{ return nConnectionID_; }
	DtnAgent*&				PeerDtnAgent()		{ return pPeerDtnAgent_; }
	TcpDtnAgent*&			TCPAgent()			{ return pTCPAgent_; }
	TcpDtnSink*&			TCPSink()			{ return pTCPSink_; }
	DtnConnection*&			NextConnection()	{ return pNextConnection_; }

	void					Connect();

protected:
	int						nConnectionID_;
	DtnAgent				*pOwnerDtnAgent_;
	DtnAgent 				*pPeerDtnAgent_;
	TcpDtnAgent				*pTCPAgent_;
	TcpDtnSink				*pTCPSink_;	
	DtnConnection			*pNextConnection_;
};

// ----------------------------------------------- ConnectionManager ---------------------------------------------- //

class ConnectionManager {
public:
	ConnectionManager(DtnAgent *pOwnerDtnAgent);
	~ConnectionManager();

	bool&					DownstreamConnectionsSet()	{ return bDownstreamConnectionsSet_; }
	void					EmptyList();
	int						DownstreamConnections();
	int						UpstreamConnections();
	void					SetupDownstreamConnections();
	void 					AddDownstreamConnection(DtnAgent *pPeerDtnAgent);
	void 					AddUpstreamConnection(DtnAgent *pPeerDtnAgent);
	bool					SetDownstreamTCP(DtnAgent *pDtnAgent, TcpDtnAgent *pTcpAgent, TcpDtnSink *pTcpSink);
	bool					SetUpstreamTCP(DtnAgent *pDtnAgent, TcpDtnAgent *pTcpAgent, TcpDtnSink *pTcpSink);
	
	DtnConnection*			GetDownstreamConnection(int nConnectionID);	
	DtnConnection*			GetUpstreamConnection(int nConnectionID);
	//DtnConnection*		GetConnection(DtnAgent *pPeerDtnAgent);
	//DtnConnection*		GetConnection(TcpDtnAgent *pSendingTCP);
	DtnConnection*			GetConnection(TcpDtnSink *pReceivingTCP);
			
protected:
	DtnAgent				*pOwnerDtnAgent_;
	DtnConnection			*pHeadDownstream_;
	DtnConnection			*pHeadUpstream_;
	bool					bDownstreamConnectionsSet_;
};

// ----------------------------------------------- DtnAgent ------------------------------------------------------ //

class DtnAgent : public Agent {

public:
	DtnAgent();
	~DtnAgent();

	virtual int				command(int argc, const char*const* argv);
	virtual void 			sendmsg(int nbytes, const char *flags = 0);
	void 					ReceiveBundlePayload(Bundle inBundle, TcpDtnSink *pTcpDtnSink);
	void					RVTimerHandler();
	void					RTTimerHandler();
	void					BundleTransmitted(Bundle inBundle);
	int						nBDProduct_;
	int						nTransferSize_;
	int						nIsBSNode_;
	int						nMaxBundleSize_;
	double					fTimeout_;
	int						nStorageSpace_;
	int						nTCPPacketSize_;

protected:
	bool					AddDtnAgent(const char* agentHandle, const char* agentType);
	bool					AttachTcp(const char* DtnAgentHandle, const char* TcpAgentHandle, const char* TcpSinkHandle, const char* agentType);
	bool					ProcessStatusReport(Bundle inBundle);
	bool					RejectBundle(Bundle inBundle);
	void					SendDataBundle(Bundle *pBundle, int nBytes);
	void					SendCustodyReport(Bundle inBundle, bool bAccept);
	void					SendOutstandingBundles();
	void					SetBundleRTTime(Bundle *pBundle);
	void 					UpdateRTTimer();
	
	ConnectionManager		connectionManager_;
	BundleManagerDTN		bundleManagerDTN_;

	int						nLastConnectionIndex_;
	int 					m_nNextBundleSeq;
	int						m_nExpectedBundleSeq;
	int						m_nExpectedSegmentSeq;
	int						m_nBytesReceived;
	double					fLastFlushTime_;
	double					fLastUtilization_;	
	
	DtnAgentRVTimer			RVtimer_; // Rendezvous
	DtnAgentRTTimer			RTtimer_; // Retransmission
	Mac802_11				*pMac_;
};

#endif
